%   File   : profile.pl
%   Author : Neng-Fa ZHOU
%   Last update : March 1997
%   Purpose: Profile programs and their executions

:- dynamic(profile_table/3).
/*
   This program can be used to analyze source programs and to gauge their 
   executions (report the numbers of predicate calls executed).

   ---------------------------------------------------------------------
   To anyalyze a program stored in one file or a list of files, use the 
   following command:

      profile_src(F) : F is the name of a file
      profile_src([F1,...,Fn]) : F1,...,Fn are names of n files

   The predicate "profile_src" reports the following information about the 
   program:
      . What predicates that are defined?
      . What predicates are used but not defined?
      , What predicates are  defined but not used?
      , What kinds of built-ins are used?

   ---------------------------------------------------------------------
   The following commands are available for gauging executions of programs.

      profile_consult(File) : same as consult(File) but with gauging ability.
      profile_compile(File) : same as consult(File) but with gauging ability.
      init_profile : initialize the profile table.
      profile : show the profile data

   To gauge the execution of a program, do the following:

       1. profile_consult(File).
          or 
          profile_compile(File),load(File).

       2. initialize the table by typing
          init_profile.

       3. execute a goal

       4. to show the profile data, type
          profile.
*/

profile_src([F|Fs]):-
    collect_res([F|Fs],Defined,Used),
    report_res(Defined,Used).
profile_src(F):-
    collect_res(F,Defined,Used),
    report_res(Defined,Used).

report_res(Defined,Used):-
    closetail(Defined),
    closetail(Used),
    builtin_preds(Used,Builtin,Used1),
    dif(Used1,Defined,Import),
    dif(Defined,Used,DefinedNotUsed),
    write(user,'sorting....'),nl(user),
    sort(Defined,DefinedSrt),
    sort(Import,ImportSrt),
    sort(Builtin,BuiltinSrt),
    sort(DefinedNotUsed,DefinedNotUsedSrt),
    prt_res(DefinedSrt,ImportSrt,DefinedNotUsedSrt,BuiltinSrt).

collect_res([],Defined,Used):-!.
collect_res(F,Defined,Used):-
    atom(F),!,
    name(F,F_l),
    name('.pl',T1),
    $append(F_l,T1,Fin_l),
    name(Fin,Fin_l),
    (exists(Fin)->
     see(Fin),
     getclauses(Clauses),
     seen,
     inf(Clauses,Used,Defined);
     warning([Fin, ' does not exist'])).
collect_res([F|Fs],Defined,Used):-
    collect_res(F,Defined,Used),
    collect_res(Fs,Defined,Used).

getclauses(Clauses):-read(X),
	(X==end_of_file->Clauses=[];
	 X=(:-op(Op,Assoc,Prec)) -> op(Op,Assoc,Prec),getclauses(Clauses);
	 Clauses=[X|Clauses1],getclauses(Clauses1)).
	
closetail(L):-var(L) : L=[].
closetail([]):-true : true.
closetail([_|Xs]):-true : closetail(Xs).

inf([],Used,Defined):-true : true.
inf([(:-B)|Cs],Used,Defined):-
	true :
	inf_body(B,Used),
	inf(Cs,Used,Defined).
inf([(H:-B)|Cs],Used,Defined):-
	true :
	functor(H,F,N),
	member1(F/N,Defined),
	inf_body(B,Used),
	inf(Cs,Used,Defined).
inf([(LHS-->RHS)|Cs],Used,Defined):-
	true :
	expand_term((LHS-->RHS),C),
	inf([C|Cs],Used,Defined).
inf([C|Cs],Used,Defined):-
	true :
	functor(C,F,N),
	member1(F/N,Defined),
	inf(Cs,Used,Defined).

inf_body(G,_):-var(G) : true.
inf_body((G : Gs),Used):- true :
	inf_body(G,Used),
	inf_body(Gs,Used).
inf_body((G ? Gs),Used):- true :
	inf_body(G,Used),
	inf_body(Gs,Used).
inf_body((G,Gs),Used):- true :
	inf_body(G,Used),
	inf_body(Gs,Used).
inf_body('->'(G,Gs),Used):-true :
	inf_body(G,Used),
	inf_body(Gs,Used).
inf_body(';'('->'(A,B),C),Used):-true :
	inf_body(A,Used),
	inf_body(B,Used),
	inf_body(C,Used).
inf_body((G;Gs),Used):-true :
	inf_body(G,Used),
	inf_body(Gs,Used).
inf_body(not(G),Used):-true :
	inf_body(G,Used).
inf_body(G,Used):- true :
	functor(G,F,N),
        member1(F/N,Used).

dif([],_,L3):-true : L3=[].
dif([X|L1],L2,L3):-
    membchk(X,L2),!,
    dif(L1,L2,L3).
dif([X|L1],L2,L3):-
	true : L3=[X|L4],
	dif(L1,L2,L4).

builtin_preds([],Builtin,Rest):-
    true :
    Rest=[],
    closetail(Builtin).
builtin_preds([Pred|Preds],Builtin,Rest):-
    F/N<=Pred ?
    predefined(F,N),!,
    member1(Pred,Builtin),
    builtin_preds(Preds,Builtin,Rest).
builtin_preds([Pred|Preds],Builtin,Rest):-
    true :
    Rest:=[Pred|Rest1],
    builtin_preds(Preds,Builtin,Rest1).

prt_res(Defined,Used,DefinedNotUsed,Builtin):-
	write('/*********************************************************************/'),nl,
	write('/* '),nl,
	write('defined:'), nl,
	p_names(Defined),
	nl,nl,
	write('used_but_not_defined:'),nl,
	p_names(Used),
        nl,nl,
	write('defined_but_not_used:'),nl,
	p_names(DefinedNotUsed),
	nl,nl,
	write('builtin:'),nl,
	p_names(Builtin),
	nl,
	write('**********************************************************************/'),nl.

p_names([]).
p_names([N]):-!,tab(5),write(N).
p_names([N|Names]):-tab(5),write(N),nl,p_names(Names).

member1(X,L):-
	var(L) : L=[X|_].
member1(X,[H|T]):-
	true : member1(X,T).

warning(Mes):-
    true :
    write(user,'** warning ** '),
    write(user,Mes),nl(user),
    fail.

%------------------------------------------------------------------------
profile_consult(File):-
    plus_ext(File,pl,InFile),
    consultable(InFile),!,
    profile_consult1(InFile).
profile_consult(File):-
    consultable(File),!,
    profile_consult1(File).

profile_consult1(InFile):-
    getclauses1(InFile,Prog,0,_),
    add_profile_clause(Prog,Prog1),
    consult_preds(Prog1,_).

profile_compile(File):-
    plus_ext(File,pl,InFile),
    consultable(InFile),!,
    plus_ext(File,out,OutFile),
    profile_compile1(InFile,OutFile).
profile_compile(File):-
    consultable(File),!,
    plus_ext(File,out,OutFile),
    profile_compile1(File,OutFile).

profile_compile1(InFile,OutFile):-
    getclauses1(InFile,Prog1,1,TabledPreds),
    add_profile_clause(Prog1,Prog2),
    phase_1_process(Prog2,Prog,TabledPreds),
    compileProgToFile(InFile,OutFile,Prog).

add_profile_clause([],NProg):-true : NProg=[].
add_profile_clause([Pred|Preds],NProg):-
    true :
    Pred=pred(Name,Arity,Mode,DB,Delay,Tabled,Cls),
    build_list(Args,Arity),
    name(Name,NameStr),
    $append("$int_",NameStr,NewNameStr),
    name(NewName,NewNameStr),
    Head=..[Name|Args],
    (Delay\==1->NPred=pred(Name,Arity,Mode,DB,Delay,Tabled,[(Head:-true ? incr_profile(NewName,Arity),fail)|Cls]);
     NPred=Pred),
    NProg=[NPred|NProg1],
    add_profile_clause(Preds,NProg1).

clear_profile :- init_profile.

init_profile :- 
    findall(X,predicate_property(X,interpreted),Xs),
    init_profile(Xs).

init_profile([]):-true : true.
init_profile([Pred|Preds]):-
    functor(Pred,Name,N),
    name(Name,NameStr),
    $append("$int_",OrigNameStr,NameStr),!,
    global_set(Name,N,0),
    init_profile(Preds).
init_profile([Pred|Preds]):-
    true :
    init_profile(Preds).
  
clear_profile_table :- retractall(profile_table(_,_,_)).

incr_profile(Name,Arity):-
    (global_get(Name,Arity,N)->
     N1 is N+1
    ;
     N1=0),
    global_set(Name,Arity,N1).

profile:-
    findall(X,predicate_property(X,interpreted),Xs),
    lookup_profile_table(Xs,Profile),
    sort(Profile,SrtProfile),
    reverse(SrtProfile,RevProfile),
    display_profile_table(RevProfile).

lookup_profile_table([],Profile):-true :
    Profile=[].
lookup_profile_table([Pred|Preds],Profile):-
    functor(Pred,F,N),
    global_get(F,N,Value),!,
    (integer(Value)->
     Profile=[profile_table(Value,F,N)|Profile1]
    ;
     Profile=Profile1),
    lookup_profile_table(Preds,Profile1).
lookup_profile_table([Pred|Preds],Profile):-
    lookup_profile_table(Preds,Profile).
    
display_profile_table([profile_table(N,Name,Arity)|Preds]):-
    true :
    name(Name,NameStr),
    ($append("$int_",OrigNameStr,NameStr) ->
     name(OrigName,OrigNameStr),
     write(OrigName/Arity),write(' : '),write(N),nl
    ;
     true),
    display_profile_table(Preds).
display_profile_table([]):-true : true.
    
build_list(L,N):-
    N=:=0 : L=[].
build_list(L,N):-
    true : 
    L=[_|L1],N1 is N-1,
    build_list(L1,N1).




    
    

    
    
